//============================================================================
// Name        : SadaCSA_Revista.cpp
// Author      : Hctor
// Version     :
// Copyright   : Your copyright notice
// Description : Hello World in C++, Ansi-style
//============================================================================
#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/resource.h>
#include "comparray4.h"
#include <limits.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

using namespace std;

#ifndef uchar
#define uchar unsigned char
#endif
#ifndef ulong
#define ulong unsigned long
#endif
#ifndef min
#define min(x,y) ((x)<(y)?(x):(y))
#endif


// Structure with all globals parameters program
typedef struct {
	uchar *seq;				// original sequence (1 byte for symbol)
	int n;				// length (numbers of symbols) of sequence

	void *index;			// LCSA index
	int samplerate;			// sampling of A

	char nameFileSeq[70];	// file's sequence name
	char typeSeq[10];		// {english, dna, proteins, sources, xml, etc}
	uint itemsExp2;			// number of continuous items for experiment 2
	uint itemsExp3;			// number of continuous items for experiment 2

	char nameDirectoryExp[150];
	double timeExp1;		// average time of experiment 1 with 'cantRepeat' repetitions
	double timeExp2;		// average time of experiment 2 with 'cantRepeat' repetitions
	double timeExp3;		// average time of experiment 3 with 'cantRepeat' repetitions
	double timeConstruction;// CPU time construction of structure (including the SA)
	bool listingTrace;		// true: print all details for console
	bool runTest;
} ParamProgram;

/* macro to detect and notify errors  */
#define IFERROR(error) {{if (error) { cout << stderr << error_index(error) << endl; exit(1); }}}
int read_file(ParamProgram *param);
void print_usage(char *);
double getTime(void);
char * error_index(int e);

double getTime (void){
	double usertime, systime;
	struct rusage usage;

	getrusage(RUSAGE_SELF, &usage);

	usertime = (double) usage.ru_utime.tv_sec +
		(double) usage.ru_utime.tv_usec / 1000000.0;

	systime = (double) usage.ru_stime.tv_sec +
		(double) usage.ru_stime.tv_usec / 1000000.0;

	return (usertime + systime);
}

/* Three function to variables to manage parameters */
static bool is_delimeter(char *delimiters, char c) {
   int i=0,len_delimiters=strlen(delimiters);
   bool is=false;
   for (i=0;i<len_delimiters;i++)
     if (c == delimiters[i]) is=true;
   return is;
}

static void parse_parameters(char *options, int *num_parameters, char ***parameters, char *delimiters) {
  int i=0,j=0,temp=0,num=0, len_options=strlen(options);
  char *options_temp;
  while  (i<len_options) {
    while ((i<len_options) && is_delimeter(delimiters,options[i])) i++;
    temp=i;
    while ((i<len_options) && !is_delimeter(delimiters,options[i])) i++;
    if (i!=temp) num++;
  }
  (*parameters) = (char **) malloc(num*sizeof(char * ));
  i=0;
  while  (i<len_options) {
    while ((i<len_options) && is_delimeter(delimiters,options[i])) i++;
    temp=i;
    while ((i<len_options) && !is_delimeter(delimiters,options[i])) i++;
    if (i!=temp) {
      (*parameters)[j]=(char *) malloc((i-temp+1)*sizeof(char));
      options_temp = options+temp;
      strncpy((*parameters)[j], options_temp, i-temp);
      ((*parameters)[j])[i-temp] = '\0';
      j++;
    }
  }
  *num_parameters = num;
}

static void free_parameters(int num_parameters,char ***parameters) {
  int i=0;
  for (i=0; i<num_parameters;i++)
    free((*parameters)[i]);
  free((*parameters));
}


/*////////////////////
//Building the Index//
////////////////////*/
extern void suffixsort(int *x, int *p, int n, int k, int l);

int load_index(char *filename, void **index){
   char fname1[128],fname2[128];
   CSA *SA;
   SA = (CSA *) malloc(sizeof(CSA));
   initranktables();
   mkdecodetable();
   sprintf(fname1,"%s.psi",filename);
   sprintf(fname2,"%s.idx",filename);
   csa_read(SA,fname1,fname2);
   (*index) = SA;
   return 0;
}

int build_index(uchar *text, int length, char *build_options, void **index, int samplerate, bool listTraza, bool runTest, double *timeConstruction, char typeSeq[10]){
  char delimiters[] = " =;";
  char filename[256];
  int j,num_parameters;
  char ** parameters;
  int rankb_w=16,rankb_w2=128;
  rankb_w = samplerate;
  int free_text=false; /* don't free text by default */
  double timeStart;

  // generate filename...
  strcpy(filename, "");
  char *str = new char[30];
  sprintf(str, "%s_n%d_%d", typeSeq, length, rankb_w);
  strcat(filename, str);
  printf("filename = (%s)\n", filename);

  if (build_options != NULL) {
    parse_parameters(build_options,&num_parameters, &parameters, delimiters);
    for (j=0; j<num_parameters;j++) {
      if ((strcmp(parameters[j], "samplerate") == 0 ) && (j < num_parameters-1) ) {
        rankb_w=atoi(parameters[j+1]);
        printf("samplerate = %d\n", rankb_w);
        j++;
      } else if  ((strcmp(parameters[j], "samplepsi") == 0 ) && (j < num_parameters-1) ) {
        rankb_w2=atoi(parameters[j+1]);
        printf("samplepsi = %d\n", rankb_w2);
        j++;
      } else if  ((strcmp(parameters[j], "filename") == 0 ) && (j < num_parameters-1) ) {
        strcpy(filename,parameters[j+1]);
        j++;
      } else if (strcmp(parameters[j], "free_text") == 0 )
        free_text=true;
    }
    free_parameters(num_parameters, &parameters);
  }

  int n=length;
  char fname1[128],fname2[128];

   /* make the SA */
  printf("make the SA\n");
   int  i,  *x, *p;
   int  k, l;
   p= (int *)malloc((n+1)*sizeof *p);
   x= (int *)malloc((n+1)*sizeof *x);
   if (! p || ! x) {
      return 1;
   }
   for ( i=0; i<n; i++) {
     x[i]=text[i];
   }
   l=0;
   k=UCHAR_MAX+1;

   suffixsort(x, p, n, k, l);
   free(x);
   p[0] = n;

   for (i=0; i<=n; ++i) p[i]++;
   if (listTraza){
	   printf("SA[1..%d] = \n",n);
	   for (i=0; i<n; i++)
		   printf("%d ",p[i]);
	   printf("\n");
   }
   // End Make SA

   timeStart = getTime();
   sprintf(fname1,"%s.psi",filename);
   sprintf(fname2,"%s.idx",filename);
   csa_new(n,p,text,fname1,fname2,rankb_w,rankb_w2);
   if (runTest == false)
	   free(p);
   if (free_text) free(text);
   load_index(filename, index);
   *timeConstruction = getTime() - timeStart;


   //************************************************* RUN TEST **********************************************
   if (runTest){
	   printf("_______________________________________________________________________________________________________\n");
	   printf("Test retrieve A[1..%d]...\n",n);

	   CSA *SA=(CSA *) *index;
	   int repetitions = 10000;
	   if (repetitions > n)
		   repetitions = n;

	   for (i=0; i<repetitions; i++){
		   j = (rand() % (n-1));
		   //printf("p[%d]=%d...\n", j, p[j]);
		   if (csa_lookup(SA, j) != p[j]){
			   printf("****** ERROR !! ******\n  csa_lookup(SA, %d) = %d != p[%d+1] = %d", j, csa_lookup(SA, j), j, p[j]);
			   exit(1);
		   }
	   }
	   printf("Test retrieve A[1..n] Finished OK !!\n");

	   printf("_______________________________________________________________________________________________________\n");
	   printf("Test retrieve A[i..j]...\n");
	   int items = 100;
	   ulong *occ = new ulong[items];
	   for (i=0; i<repetitions; i++){
		   j = (rand() % (n-items+1));
		   occ = csa_batchlookup2(SA,j,j+items-1);
		   for (k=0; k<10; k++){
		      //printf("p[%d]=%d...\n", j, p[j]);
			   if (occ[k] != (p[j+k]-1)){
				   printf("****** ERROR !! ******\n  occ[%d] = %d != p[%d+%d] = %d", k, occ[k], j, k, p[j+k]);
				   exit(1);
			   }
		   }
	   }

	   printf("Test retrieve A[i..j] Finished OK !!\n");
   	   free(p);
   }

   return 0;
}

int save_index(void *index, char *filename) {
  return 99;
}

int free_index(void *index){
  CSA *SA=(CSA *) index;
  #ifdef USE_MMAP
  if(SA->mapp!=NULL)
		free(SA->mapp);
  if(SA->mapi!=NULL)
		free(SA->mapi);
  #else
  free(SA->B)
  free(SA->ISA);
  free(SA->R);
  free(SA->SA);
  #endif
  free(SA);
  return 0;
}
int index_size(void *index, ulong *size) {
  CSA *SA=(CSA *) index;
  *size=0;
  #ifdef USE_MMAP
  *size+=sizeof(MMAP)*2;
  *size+=(SA->p_size); //Size B
  *size+=(SA->i_size); //Size ISA+R+SA
  #endif
  *size+=sizeof(CSA);
  return 0;
}

int index_size_count(void *index, ulong *size) {
  CSA *SA=(CSA *) index;
  *size=0;
  #ifdef USE_MMAP
  *size+=sizeof(MMAP)*2;
  *size+=(SA->p_size); //Size B
  *size+=(SA->i_size); //Size ISA+R+SA
  #endif
  *size+=sizeof(CSA);
  return 0;
}
int index_size_locate(void *index, ulong *size) {
  CSA *SA=(CSA *) index;
  *size=0;
  #ifdef USE_MMAP
  *size+=sizeof(MMAP)*2;
  *size+=(SA->p_size); //Size B
  *size+=(SA->i_size); //Size ISA+R+SA
  #endif
  *size+=sizeof(CSA);
  return 0;
}

/*////////////////////
//Querying the Index//
////////////////////*/
int count(void *index, uchar *pattern, ulong length, ulong *numocc){
  int l,r,len;
  CSA *SA=(CSA *) index;
  len = csa_bsearch(pattern,length,SA,&l,&r);
  *numocc = r-l+1;
  return 0;
}

int get_length(void *index, ulong *length){
   (*length)=((CSA *) index)->n;
   return 0;
}

/*////////////////
//Error handling//
////////////////*/
char * error_index(int e){
  char *error = new char[50];
	switch(e) {
    case 0:  printf(error, "No error"); break;
    case 1:  printf(error, "Out of memory"); break;
    case 2:  printf(error, "The text must end with a \\0"); break;
    case 5:  printf(error, "You can't free the text if you don't copy it"); break;
    case 20: printf(error, "Cannot create files"); break;
    case 21: printf(error, "Error writing the index"); break;
    case 22: printf(error, "Error writing the index"); break;
    case 23: printf(error, "Cannot open index; break");
    case 24: printf(error, "Cannot open text; break");
    case 25: printf(error, "Error reading the index"); break;
    case 26: printf(error, "Error reading the index"); break;
    case 27: printf(error, "Error reading the text"); break;
    case 28: printf(error, "Error reading the text"); break;
    case 99: printf(error, "Not implemented"); break;
    default: printf(error, "Unknown error");
  }
	return error;
}

//*************************************************************************************************
//  Opens and reads a text file
int read_file(ParamProgram *param) {
  uchar *text;
  unsigned long t;
  FILE *infile;

  infile = fopen(param->nameFileSeq, "rb"); // rb is for binary: required by DOS
  if(infile == NULL) return 1;

  // store input file length
  if(fseek(infile,0,SEEK_END) !=0 ) return 1;
  param->n = ftell(infile);

  // alloc memory for text (the overshoot is for suffix sorting)
  text = (uchar *) malloc(param->n*sizeof(*text));
  if(text == NULL) return 1;

  cout << "Reading sequence from original file, input_size = " << param->n << endl;
  // read text in one sweep
  rewind(infile);
  t = fread(text, sizeof(*text), (size_t) param->n, infile);
  if(t!=param->n) return 1;
  param->seq = text;
  fclose(infile);

  cout << "file read" << endl;

  return 0;
}

// compute ceiling to logarithm base 2 for num
uint computeCeilingLogNum(ulong num){
	uint lg = 0;
	float lgB = log(num)/log(2);

	if((lgB - (uint)lgB) > 0)
    	lg = (uint)lgB + 1;
    else
    	lg = (uint)lgB;

	return lg;
}

// Experiment 1 - compute A[i]: it gives the average time of CPU
double runExperimentOne(ParamProgram *param, uint repetitions){
	double t, avgTime = 0.0;
	unsigned i, j, value;
	CSA *SA=(CSA *) param->index;

	printf("Start Experiment 1,  Extract A[i], for %d repetitions\n",repetitions);

	for (i=0; i<repetitions; i++){
		j = (rand() % param->n);
		t = getTime();
		value = csa_lookup(SA, j);
		t = getTime() - t;
		avgTime += t/(double)repetitions;
		//printf("A[%d] = %d\n",j,value);
	}

	printf("Average CPU time for item : %f Milliseconds\n", avgTime * 1000.0);
	printf("Experiment 1 has finished\n");
	printf("_______________________________________________________________________________________________________\n");

	return avgTime;
}

// Experiment 2 - compute A[i..j]: it gives the average time of CPU
double runExperimentTwo(ParamProgram *param, uint itemsExp2, uint repetitions, unsigned long *values){
	double t, avgTime = 0.0;
	unsigned i, j;
	CSA *SA=(CSA *) param->index;

	printf("Start Experiment 2. Extract A[i..i+%d], for %d repetitions\n", itemsExp2, repetitions);

	for (i=0; i<repetitions; i++){
		j = (rand() % (param->n - itemsExp2 + 1));
		t = getTime();
		values = csa_batchlookup2(SA,j,j+itemsExp2-1);
		t = getTime() - t;
		avgTime += t/(double)repetitions;
	}
	t = getTime() - t;
	printf("Average CPU time for item : %f Milliseconds\n", avgTime * 1000.0);
	printf("Experiment 2 has finished\n");
	printf("_______________________________________________________________________________________________________\n");

	return avgTime;
}

// Experiment 3 - compute A[i..i+1], A[i..i+2], ..., A[i..i+%d], it gives the average time of CPU
double runExperimentThree(ParamProgram *param, double *timesItems, uint repetitions, unsigned long *values){
	double t;
	unsigned r, i, j;
	CSA *SA=(CSA *) param->index;

	for (i=0; i<param->itemsExp3; i++)
		timesItems[i] = 0.0;

	printf("Start Experiment 3. Extract A[i], A[i..i+1], A[i..i+2], ..., A[i..i+%d], for %d repetitions\n", param->itemsExp3-1, repetitions);
	for (i=0; i<param->itemsExp3; i++){
		for (r=0; r<repetitions; r++){
			j = (rand() % (param->n - param->itemsExp3 + 1));
			t = getTime();
			values = csa_batchlookup2(SA,j,j+i);
			t = getTime() - t;
			timesItems[i] = timesItems[i] + t/(double)repetitions;
		}
	}

	printf("CPU time for each c = {1,2,...,%d} in Milliseconds\n", param->itemsExp3);
	for (i=0; i<param->itemsExp3; i++)
		printf("%f ", timesItems[i]*1000.0);

	printf("\nExperiment 3 has finished\n");
	printf("_______________________________________________________________________________________________________\n");

	return 0;
}

// Run experiments and create result file
void runExperiments(ParamProgram *param){
	unsigned long *values = new unsigned long[100];
	double *timesItems = new double[param->itemsExp3];

	uint rep = 1000000;

	param->timeExp1 = runExperimentOne(param, rep);
	param->timeExp2 = runExperimentTwo(param, param->itemsExp2, rep, values);
	if (param->n > 200)
		param->timeExp2 = runExperimentTwo(param, 100, rep, values);
	param->timeExp3 = runExperimentThree(param, timesItems, rep, values);

	delete (unsigned *) values;
	delete (double *) timesItems;
}

void print_usage(char * progname) {
	cout << "ERRORR !! " << endl;
	cout << "SadCSA's usage: " << endl;
	cout << " " << progname << " <source file> <samplerate>" << endl;
	exit(1);
}


int main(int argc, char *argv[]) {
	char *params = NULL;
	int error;

	ParamProgram param;
	param.listingTrace = false;
	param.runTest = false;
	param.itemsExp2 = 10;
	param.itemsExp3 = 20;
	param.samplerate = 16; //default value in original code

	if(argc < 4) print_usage(argv[0]);
	strcpy(param.nameFileSeq, argv[1]);
	param.samplerate = atoi(argv[2]);
	strcpy(param.typeSeq, argv[3]);
	cout << "SadCSA's Parameters..." << endl;
	cout << "File: " << param.nameFileSeq << endl;
	cout << "samplerate: " << param.samplerate << endl;
	cout << "Type Seq: " << param.typeSeq << endl;

	error = read_file(&param);
	IFERROR(error);
	if(param.listingTrace) {
		cout <<"Text:" << endl;
		cout << param.seq << endl;
	}

	cout << "Create index..." << endl;
	error = build_index(param.seq, param.n, params, &(param.index), param.samplerate, param.listingTrace, param.runTest, &(param.timeConstruction), param.typeSeq);
	IFERROR(error);

	ulong length =  0;
	get_length(param.index, &length);
	cout << "Index length: " << length << endl;
	cout << "++++++++++++++++++++++++++++++++++++++++++" << endl;
	cout << "CPU time construction: " << param.timeConstruction * 1000.0 << " Milliseconds" << endl;
	cout << "++++++++++++++++++++++++++++++++++++++++++" << endl;
	cout << endl << "Index created" << endl;

	ulong cantBytesCSA = 0;
	error = index_size(param.index, &cantBytesCSA);
	cout << "++++++++++++++++++++++++++++++++++++++++++" << endl;
	cout << "Index size = " << cantBytesCSA << " Bytes" << ", " << (float)cantBytesCSA/(float)(param.n) << "|Text|" << endl;
	cout << "++++++++++++++++++++++++++++++++++++++++++" << endl;

	// Run experiments...
	srand(time(NULL));
	runExperiments(&param);

	cout << "Program Finished OK" << endl;
	cout << "_______________________________________________________________________________________________________" << endl;
	cout << "$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$" << endl << endl;
	return 0;
}
